// vim:noexpandtab:tabstop=4:softtabstop=4:shiftwidth=4
/**
* Test MPI and Spawn_multiple on the SGE grid.
*/
using namespace std;
#include <cstdio>
#include <cstdlib>
extern "C" {
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <pwd.h>
}
#include <Mpitest.hpp>


Mpitest::
Mpitest( int argc, char *argv[], char *envp[] ) throw( MpitestRunException )
{
	_progname.assign( basename( argv[ 0 ] ) );
	_debug = false;
	_root = true;	// assume we are the root for now.
	_sleep = 10;	// default root sleep for 2 seconds.
	_nProc = 4;
	_command = 0;
	_arg = 0;
	_maxProc = 0;
	_info = 0;

	int		c;
	optind = 0;	// extern in getopt.h
	opterr = 0; // extern in getopt.h

	// Process the command line arguments.
	for(;;) {
		static struct option long_options[] = {
			{ "debug",				no_argument,		0,	'd' },
			{ "child",				no_argument,		0,	'c' },
			{ "np",                 required_argument,	0,	'n' },
			{ "sleep",				required_argument,	0,	's' },
			{ 0, 0, 0, 0 }
		};
		int option_index = 0;

		c = getopt_long( argc, argv, "dcn:s:",
		  long_options, &option_index );

		if( c == -1 ) break;

		switch( c ) {

		case 'd':	// --debug
			_debug = true;
			break;

		case 'c':	// --child
			_root = false;
			break;

		case 'n':	// --np N  (number of child processes)
			_nProc = strtol( optarg, 0, 0 );
			break;

		case 's':	// --sleep seconds
			_sleep = strtol( optarg, 0, 0 );
			break;

		default:
			usage();
			break;
		}
	} // for(;;)
	
	// Optional non-flag arguments.
	switch( argc - optind ) {

	case 0:
		break;

	case 1:
		usage();
		break;

	default:
		usage();
		break;
	}

	// Initialize the MPI system.
	if( MPI::Init_thread( MPI::THREAD_MULTIPLE ) != MPI::THREAD_MULTIPLE ) {
		throw MpitestRunException(
		  string( "MPI library does not support MPI::THREAD_MULTIPLE." ) );
	}

	// Root process will overwrite _intercomm once it spawns it's
	// children to represent the intercommunicator used with the children.
	_intercomm = MPI::COMM_WORLD.Get_parent();
	_rank = MPI::COMM_WORLD.Get_rank();

	if( _root == true ) {
		// root (parent) process
		if( _rank != 0 ) {
			throw MpitestRunException( string(
			  "Root process must have rank 0, got rank: "
			  + stringify( _rank ) ) );
		}
		if( _intercomm != MPI::COMM_NULL ) {
			throw MpitestRunException( string(
			  "Root process should not have a MPI parent but does." ) ) ;
		}
		MPI::COMM_WORLD.Set_errhandler( MPI::ERRORS_THROW_EXCEPTIONS );
	} else {
		// child process
		_intercomm.Set_errhandler( MPI::ERRORS_THROW_EXCEPTIONS );

		if( _intercomm == MPI::COMM_NULL ) {
			throw MpitestRunException( string(
			  "Child process does not have a MPI parent, rank: "
			  + stringify( _rank ) ) );
		}
	}
}


/** Destructor.
*/
Mpitest::
~Mpitest()
{
	for( int rank = 0;  rank < _nProc;  rank++ ) {

		if( _command != 0 && _command[ rank ] != 0 )
			delete _command[ rank ];

		if( _arg != 0 && _arg[ rank ] != 0 ) {
			for( char **p = _arg[ rank ];  *p;  p++ ) {
			  delete *p;
			}
			delete _arg[ rank ];
		}
		if( _info != 0 && _info[ rank ] != 0 ) {
			_info[ rank ].Free();
		}
	}

	if( _command != 0 )
		delete _command;

	if( _arg != 0 )
		delete _arg;

	if( _maxProc != 0 )
		delete _maxProc;

	Mpitest::finalize();
}


void Mpitest::
run() throw( MpitestRunException, MPI::Exception )
{
	if( _root == true ) {
		childrenSpawn();	// No tokophobia here!
		rootTest();		// private local routine
	} else {
		childTest();	// private local routine
	}
}


/** Spawn children processes.
*/
void Mpitest::
childrenSpawn() throw( MpitestRunException, MPI::Exception )
{
	int		rank, argc;
	char	**argv;
	string	wdir, command;;

	cout << "Entering childrenSpawn" << endl;
	cout << _progname << ": Starting " << _nProc;
	if( _nProc == 1 )
		cout << " child process.";
	else
		cout << " children processes.";
	cout << endl;
	_command = new char*[ _nProc ];
	_arg = new char **[ _nProc ];
	memset( _arg, 0, _nProc * sizeof( char ** ) );
	_maxProc = new int[ _nProc ];
	_info = new MPI::Info[ _nProc ];
	string exe( _progname );
	which( exe );	// locate "this" executable in PATH

	// Build command for spawn-multiple.
	for( rank = 0;  rank < _nProc;  rank++ ) {
		_maxProc[ rank ] = 1;
		_info[ rank ] = MPI::Info::Create();

		// NOTA BENE - openmpi does not implement the "path" MPI_Info
		// reserved key.  Further, openmpi interprets "wdir" as the location
		// of the executable rather than as a separate "working directory"
		// as the MPI2 specification implies. 
		// However, "wdir" does not seem to be interpreted correctly either
		// so create the "command" as the full path to the executable.
		//
		// The "wdir" reserved key is the path to "this" executable.
		wdir = exe;
		argv = new char*[ 15 ];	// Nota Bene fixed array length.
		argc = 0;
		_arg[ rank ] = argv;

		argv[ argc++ ] = strdup( "--child" );

		if( _debug == true ) {
			argv[ argc++ ] = strdup( "--debug" );
		}

		argv[ argc++ ] = 0;	// terminate argv array

		// Insert MPI::Info key/value pairs.
		_info[ rank ].Set( "wdir", wdir.c_str() );

		command = wdir + "/" + _progname;
		_command[ rank ] = strdup( command.c_str() );

		if( access( _command[ rank ], X_OK ) != 0 ) {
			throw MpitestRunException( string( 
			  "Cannot access or no execute privileges on: " + command ) );
		}

		// If debug flag set print the argv information.
		if( _debug == true ) {
			string cli( command );;
			char **	argvp = argv;

			while( *argvp != 0 ) {
				cli += " ";
				cli += *argvp++;
			}
			cout << "Rank " <<  rank << " CLI: " << cli << endl;
		}
	}

	int *errCode = new int[ _nProc ];

	cout << "Root calling spawn multiple" << endl;
	// Spawn all of the children processes.
	_intercomm = MPI::COMM_WORLD.Spawn_multiple( _nProc,
	  const_cast<const char **>(_command),
	  const_cast<const char ***>(_arg),
	  _maxProc, _info, 0, errCode );
	cout << "Root returned from spawn multiple call." << endl;

	// Errors-throw-exceptions was already set on the root's
	// COMM_WORLD, set it on the root/child intercommunicator as well.
	_intercomm.Set_errhandler( MPI::ERRORS_THROW_EXCEPTIONS );

	// Check all of the children return codes.
	int nFail = 0;
	for( rank = 0;  rank < _nProc;  rank++ ) {
		if( errCode[ rank ] != MPI_SUCCESS ) {
			nFail++;
		}
	}
	if( nFail > 0 ) {
		throw MpitestRunException( stringify( nFail )
		  + " of " + stringify( _nProc )
		  + " children failed to spawn." );
	}
	cout << "Leaving childrenSpawn" << endl;
	delete errCode;
}


/** Run root test.
* Receive a message from all the "test" children and once all
* have responded send them all a message telling them to exit.
* Note that if one or more of the children fail to respond, this
* test just hangs waiting for the missing response(s).
* @throw MPI::Exception MPI system indicating error condition.
*/
void Mpitest::
rootTest() throw( MPI::Exception )
{
	MPI::Status status;
	char buf[ 32 ];

	cout << "Entering rootTest" << endl;
	for( int nProc = 0;  nProc < _nProc;  nProc++ ) {

		cout << "Root calling Recv." << endl;
		_intercomm.Recv( buf, sizeof( buf ), MPI::CHAR,
		  MPI::ANY_SOURCE, MPI::ANY_TAG, status );

		cout << "Root rcv from rank " << status.Get_source() << ": "
		 << buf << endl;
	}

	cout << "Root sleeping for " << stringify( _sleep ) << " seconds." << endl;
	sleep( _sleep );

	strncpy( buf, "Exit", sizeof( buf ) );
	int count = strlen( buf ) + 1;;

	for( int nProc = 0;  nProc < _nProc;  nProc++ ) {
		cout << "Root calling Send." << endl;
		_intercomm.Send( buf, count, MPI::CHAR, nProc, 1234 );
	}
	sleep( 2 );
	cout << "Leaving rootTest" << endl;
}


/** Child test.
* Send a message to the root's test method and wait for a response.
* If the response message has an MPI "tag" of 1234, return.
* Otherwise wait for another message.
* @throw MPI::Exception MPI system indicating error condition.
*/
void Mpitest::
childTest() throw( MPI::Exception )
{
	MPI::Status status;
	char buf[ 32 ];

	cout << "Entering childTest" << endl;
	strncpy( buf, "Hello", sizeof( buf ) );
	int count = strlen( buf ) + 1;

	cout << "Rank " << MPI::COMM_WORLD.Get_rank() 
	  << " sleeping for " << stringify( _sleep ) << " seconds." << endl;
	sleep( _sleep );

	cout << "Rank " << MPI::COMM_WORLD.Get_rank() 
	     << " calling Send." << endl;
	_intercomm.Send( buf, count, MPI::CHAR, 0, 0 );

	do {
		cout << "Rank " << MPI::COMM_WORLD.Get_rank() << " calling Recv." << endl;
		_intercomm.Recv( buf, sizeof( buf ), MPI::CHAR,
		  MPI::ANY_SOURCE, MPI::ANY_TAG, status );

		cout << "Rank " << MPI::COMM_WORLD.Get_rank()
		  << " from Root: " << buf << endl;
	} while( status.Get_tag() != 1234 );
	cout << "Leaving childTest" << endl;
}


/** Throw exception with usage message.
* @throw MpitestRunException
*/
void Mpitest::
usage() throw( MpitestRunException )
{
	throw( MpitestRunException( string( "Usage: progname executable [options]\n"
	"\toptions:\n"
	"\t[--debug|-d]\n"
	"\t[--child|-c]\n"
	) ) );
}

/** Routine to finalize the MPI system.
*/
void Mpitest::
finalize()
{
	if( MPI::Is_finalized() == false ) {
		MPI::Finalize();
	}
}


/** Discover the location of specified executable based on current
* environment's PATH parameter.
* The notion of the current working directory is always checked even
* if "." is not part of the PATH environment (checked last when "."
* not present).
* @param progpath On input progpath is the name of the program to
*  locate, on return progpath is replaced with the path to the
*  executable program which does not include the program name
*  (in other words, basename(progpath) != the input progpath),
*  progpath will be the empty string if a path could not be found.
*/
void Mpitest::
which( string& progpath )
{
	char buf[ 1024 ];	// working character buffer
	string cwd( getcwd( buf, sizeof( buf ) ) );
	string path( getenv( "PATH" ) );
	path.append( ":" + cwd + ":" ); // NB trailing colon
	vector<string> paths;
	size_t pos1 = 0, pos2;

	while( (pos2 = path.find_first_of( ':', pos1 )) != string::npos ) {
		// Leading ':' or perhaps two ':' chars in a row => pos1 == pos2.
		if( pos1 != pos2 ) {
			string dir( path.substr( pos1, pos2 - pos1 ) );
			if( dir == "." && cwd.empty() == false ) {
				paths.push_back( cwd );
			} else {
				paths.push_back( dir );
			}
		}
		pos1 = pos2 + 1;
	}

	for( vector<string>::iterator it = paths.begin();
	  it < paths.end();  it++ ) {

		string command( *it + "/" + progpath );

		if( access( command.c_str(), X_OK ) == 0 ) {
			progpath.assign( *it );
			return;
		}
	}
	progpath.clear();
}



int
main( int argc, char **argv, char **envp ) {
	try {
		Mpitest *mpitest = new Mpitest( argc, argv, envp );
		mpitest->run();
		delete mpitest;
	} catch( MpitestRunException e ) {
		cerr << e.what() << endl;
		return 1;
	} catch( MPI::Exception e ) {
		cerr << "MPI EXCEPTION: " << e.Get_error_string() << endl;
		return 2;
	}
	return 0;
}
