// vim:noexpandtab:tabstop=4:softtabstop=4:shiftwidth=4
/**
* @file Mpitest.hpp
* @brief A test for MPI tight integration with SGE and Spawn_multiple.
*/

#ifndef _Mpitest_hpp_
#define _Mpitest_hpp_
#include <string>
#include <vector>
#include <map>
#include <fstream>
#include <sstream>
#include <iostream>
#include <typeinfo>
#include <stdexcept>
#include <mpi.h>

extern "C" {
#include <sys/types.h>
}


/** Exception class returned by stringify() for failed conversion.
*/
class BadConversion : public std::runtime_error {
  public:
	/** BadConversion constructor.
	* @param s String to pass to stdexcept runtime_error().
	*/
	BadConversion( const std::string& s ) : std::runtime_error( s ) {}
};


/** Template function for converting objects to strings.
* @param x Object supporting iostream or fundamental type.
* @return Reference to string representing x.
*/
template <typename T> inline std::string
stringify( const T& x )
{
	std::ostringstream o;
	if( !(o << x) )
		throw BadConversion( std::string( "stringify(" )
		  + typeid(x).name() + ")" );
	return o.str();
}

/** Template class for converting objects to strings.
* Be smart and just return reference to already existing string.
* @param t String reference.
* @return Reference to string representing x.
*/
template <> inline std::string
stringify<string>(const std::string& t)
{
	return t;
}


class MpitestRunException : public std::runtime_error {
 public:
	MpitestRunException( const std::string& s ) : std::runtime_error( s ) {}
};


/** Class Mpitest.
*/
class Mpitest {
 public:
	static const int RankMax = 2048;/**< max no. of MPI spawned processes */
 private:
	bool	_debug;					/**< true => debug output */
	bool	_root;					/**< true => root process, false => child */
	int		_rank;					/**< MPI process rank */
	long int _sleep;				/**< sleep time */
	string	_progname;				/**< invocation name of process */
	MPI::Intercomm	_intercomm;		/**< intercommunicator parent/children */

	// Command line creation -- kept here so the destuctor can clean up.
	long int _nProc;					/**< # entries in following arrays */
	char	**_command;				/**< array of executable paths */
	char	***_arg;				/**< array of pointers to arrays of args */
	int		*_maxProc;				/**< array of max proc (all entries 1) */
	MPI::Info	*_info;				/**< see MPI::Comm::spawn_multiple */

	// Private class methods.
	void childrenSpawn() throw( MpitestRunException, MPI::Exception );
	void rootTest() throw( MPI::Exception );
	void childTest() throw( MPI::Exception );
	void usage() throw( MpitestRunException );
	static void finalize();

 public:
	Mpitest( int argc, char *argv[], char *envp[] )
	  throw( MpitestRunException );
	~Mpitest();

	/** @return Program invocation name.
	*/
	const string& progname() { return _progname; }

	/** @return debug flag.
	*/
	bool debug() { return _debug; }

	/** @return MPI rank.
	*/
	int rank() { return _rank; }

	void run()
	  throw( MpitestRunException, MPI::Exception );

	void which( string& progpath );
};

#endif
